@@ -1,4 +1,4 @@
-from os import environ
+from os import environ, exit
 import logging
 from socket import getaddrinfo, gaierror
 from time import sleep
@@ -35,6 +35,21 @@ def wait_for_dns(domain, port):
             sleep(1)
 
 
+def back_off(n_times, func):
+    i = 0
+    while True:
+        i += 1
+        if i > n_times:
+            logger.warning("Cannot resolve Gimulator host! Retrying...")
+            exit(1)
+        
+        try:
+            return func()
+        except Exception as e:
+            logger.warning("Cannot resolve Gimulator host! Retrying...")
+        sleep(1)
+
+
 class GimulatorClient:
     domain = None
     port = None
@@ -62,27 +77,27 @@ class GimulatorClient:
 
     def Get(self, key: Key):
         wait_for_dns(self.domain, self.port)  # FIXME this is irritating! Someone please fix this...
-        return self.api.Get(key, metadata=self.metadata)
+        return back_off(5, lambda: self.api.Get(key, metadata=self.metadata))
 
     def GetAll(self, key: Key):
         wait_for_dns(self.domain, self.port)  # FIXME this is irritating! Someone please fix this...
-        return self.api.GetAll(key, metadata=self.metadata)
+        return back_off(5, lambda: self.api.GetAll(key, metadata=self.metadata))
 
     def Put(self, message: Message):
         wait_for_dns(self.domain, self.port)  # FIXME this is irritating! Someone please fix this...
-        return self.api.Put(message, metadata=self.metadata)
+        return back_off(5, lambda: self.api.Put(message, metadata=self.metadata))
 
     def Delete(self, message: Message):
         wait_for_dns(self.domain, self.port)  # FIXME this is irritating! Someone please fix this...
-        return self.api.Delete(message, metadata=self.metadata)
+        return back_off(5, lambda: self.api.Delete(message, metadata=self.metadata))
 
     def DeleteAll(self, message: Message):
         wait_for_dns(self.domain, self.port)  # FIXME this is irritating! Someone please fix this...
-        return self.api.DeleteAll(message, metadata=self.metadata)
+        return back_off(5, lambda: self.api.DeleteAll(message, metadata=self.metadata))
 
     def Watch(self, key: Key):
         wait_for_dns(self.domain, self.port)  # FIXME this is irritating! Someone please fix this...
-        return self.api.Watch(key, metadata=self.metadata)
+        return back_off(5, lambda: self.api.Watch(key, metadata=self.metadata))
 
 
 class DirectorClient(GimulatorClient):
@@ -100,28 +115,21 @@ class DirectorClient(GimulatorClient):
 
             try:
                 users = list(self.GetActors())
-                if len(users) == 1:
+                if len(users) > 0:
                     return
             except grpc._channel._MultiThreadedRendezvous:  # FIXME this is irritating! Someone please fix this...
                 logger.warning("Cannot resolve Gimulator host! Retrying...")
                 sleep(1)
+            except Exception as e:
+                logger.warning("unknown grpc error: %s" % e)
+                sleep(1)
 
     def GetActors(self):
         empty = google_dot_protobuf_dot_empty__pb2.Empty()
-        while True:
-            try:
-                return self.director_api.GetActors(empty, metadata=self.metadata)
-            except grpc._channel._MultiThreadedRendezvous:  # FIXME this is irritating! Someone please fix this...
-                logger.warning("Cannot resolve Gimulator host! Retrying...")
-                sleep(1)
+        return back_off(5, lambda: self.director_api.GetActors(empty, metadata=self.metadata))
 
     def PutResult(self, result: Result):
-        while True:
-            try:
-                return self.director_api.PutResult(result, metadata=self.metadata)
-            except grpc._channel._MultiThreadedRendezvous:  # FIXME this is irritating! Someone please fix this...
-                logger.warning("Cannot resolve Gimulator host! Retrying...")
-                sleep(1)
+        return back_off(5, lambda: self.director_api.PutResult(result, metadata=self.metadata))
 
 
 class OperatorClient(GimulatorClient):
@@ -133,7 +141,7 @@ class OperatorClient(GimulatorClient):
 
     def SetUserStatus(self, report: Report):
         wait_for_dns(self.domain, self.port)  # FIXME this is irritating! Someone please fix this...
-        return self.operator_api.SetUserStatus(report, metadata=self.metadata)
+        return back_off(5, lambda: self.operator_api.SetUserStatus(report, metadata=self.metadata))
 
 
 class ActorClient(GimulatorClient):
@@ -146,4 +154,4 @@ class ActorClient(GimulatorClient):
     def ImReady(self):
         empty = google_dot_protobuf_dot_empty__pb2.Empty()
         wait_for_dns(self.domain, self.port)  # FIXME this is irritating! Someone please fix this...
-        return self.actor_api.ImReady(empty, metadata=self.metadata)
+        return back_off(5, lambda: self.actor_api.ImReady(empty, metadata=self.metadata))
